// @ts-check
const fs = require(`fs`);
const Canvas = require(`canvas`);
const xml2js = require(`xml2js`);

const resize = 10;

/**
 * 
 * @param {Canvas.Canvas} c 
 * @returns 
 */
function trimCanvas(c) {
    var ctx = c.getContext('2d');
    var pixels = ctx.getImageData(0, 0, c.width, c.height);
    var l = pixels.data.length,
        i,
        bound = {
            top: null,
            left: null,
            right: null,
            bottom: null
        },
        x, y;

    // Iterate over every pixel to find the highest
    // and where it ends on every axis ()
    for (i = 0; i < l; i += 4) {
        if (pixels.data[i + 3] === 0) continue;
        x = (i / 4) % c.width;
        y = ~~((i / 4) / c.width);

        if (bound.top === null) {
            bound.top = y;
        }

        if (bound.left === null) {
            bound.left = x;
        } else if (x < bound.left) {
            bound.left = x;
        }

        if (bound.right === null) {
            bound.right = x;
        } else if (bound.right < x) {
            bound.right = x;
        }

        if (bound.bottom === null) {
            bound.bottom = y;
        } else if (bound.bottom < y) {
            bound.bottom = y;
        }
    }

    // Calculate the height and width of the content
    let trimHeight = bound.bottom - bound.top;
    let trimWidth = bound.right - bound.left;
    if (!trimWidth && !trimHeight) return;
    let trimmed = ctx.getImageData(bound.left, bound.top, trimWidth, trimHeight);

    // copy
    ctx.canvas.width = trimWidth;
    ctx.canvas.height = trimHeight;
    ctx.putImageData(trimmed, 0, 0);

    // Return trimmed canvas
    ctx.imageSmoothingEnabled = false;
    return { canvas: ctx.canvas, bound };
}

/**
 * 
 * @param {import('./meta.json')} meta 
 */
async function make(meta) {

    const skinXml = `../src/main/res/xml/skin.xml`
    //let out = await xml2js.parseStringPromise(fs.readFileSync(skinXml));
    //console.debug(JSON.stringify(out, null, 2));
    const states = [`stop`, `wait`, `awake`, `moveUp`, `moveDown`, `moveLeft`, `moveRight`, `moveUpLeft`, `moveUpRight`, `moveDownLeft`, `moveDownRight`/*, `wallLeft`, `wallRight`*/];
    const skin = {
        "motion-params": {
            "$": {
                "acceleration": 200,
                "maxVelocity": 120,
                "deaccelerationDistance": 60,
                "proximityDistance": 10,
                "initialState": "stop",
                "awakeState": "awake",
                "moveStatePrefix": "move",
                "wallStatePrefix": "wall"
            },
            "motion": states.map(state => {
                let motion = {
                    $: { state },
                };
                switch (state) {
                    case 'stop':
                        motion.$.nextState = `wait`;
                        motion.$.checkWall = `true`;
                        break;
                    case 'awake':
                        motion.$.nextState = `stop`;
                        motion.$.checkMove = `true`;
                        break;
                }
                if (state.startsWith(`wall`)) {
                    motion.$.nextState = `wait`;
                    motion.$.repeatCount = `3`;
                }
                return motion;
            })
        }
    };
    function addMotion(state, drawable, duration) {
        let motion = skin['motion-params'].motion.find(x => x['$'].state === state);
        if (!motion) throw new Error(`couldn't find motion for state: ${state}`);
        const item = {
            $: {
                drawable: `@drawable/${drawable}`,
                duration: duration * 1000 || 400,
            }
        }
        const noRepeat = [`stop`, 'awake'].includes(state);
        // @ts-ignore
        if (motion["repeat-item"]) return motion["repeat-item"][0].item.push(item);
        if (motion.item) {
            if (noRepeat) {
                if (!Array.isArray(motion.item))
                    motion.item = [motion.item];
                motion.item.push(item);
                return;
            }
            motion["repeat-item"] = [{ item: [] }];
            motion["repeat-item"][0].item.push(motion.item);
            motion["repeat-item"][0].item.push(item);
            delete motion.item;
            return;
        }
        // @ts-ignore
        motion.item = item;
    }

    //fs.rmSync(`./out`, {force:true,recursive:true})
    //console.log(`v${meta.version} - ${meta.license} - ${meta.copyright}`);
    const outDir = `../src/main/res/drawable/`;
    if (!fs.existsSync(outDir)) fs.mkdirSync(outDir);
    const totalSize = meta.size;
    const name = "rainbow_baby_slime";
    const state = meta.states.find(x => x.name === name);
    if (!state) throw new Error(`AAAAAAAAAAAAAA NO ${name} STATE`);
    if (!state.delays) throw new Error(`No delays TODO size based`);

    function getFrames(path, bound = { top: 0, left: 0, }) {
        console.debug(path)
        let img = new Canvas.Image;
        img.onerror = console.error
        img.src = fs.readFileSync(path);
        const total = (img.height / totalSize.x) * (img.width / totalSize.y);
        let frames = [];
        for (let i = 0; i < total; i++) {
            const floor = Math.floor(i / (img.width / totalSize.y));
            const frame = i - Math.floor(floor * (img.width / totalSize.y))
            const x = totalSize.x * frame;
            const y = totalSize.y * floor;
            frames.push(ctx => {
                ctx.drawImage(img,
                    x, y,
                    totalSize.x, totalSize.y,
                    0, 0,
                    totalSize.x * resize, totalSize.y * resize
                )
            })
        }

        return frames;
    }

    const bodyFrames = getFrames(`${name}.png`);
    const faceFrames = getFrames(`aslime-_3.png`);

    for (let bi = 0; bi < bodyFrames.length - 2; bi++) {

        for (let fi = 0; fi < 4; fi++) {
            var canvas = Canvas.createCanvas(totalSize.x * resize, totalSize.y * resize);
            let ctx = canvas.getContext('2d');
            ctx.imageSmoothingEnabled = false;

            // body
            if (bodyFrames[bi])
                bodyFrames[bi](ctx);
            //let trim = trimCanvas(canvas);
            //if (trim) {
            //canvas = trim.canvas;
            ctx = canvas.getContext('2d');

            // face
            if (faceFrames[fi])
                faceFrames[fi](ctx/*, trim.bound*/);
            //}

            // output
            let output = [];
            switch (fi) {
                case 0:
                    output.push(`${outDir}down${bi}.png`);
                    addMotion(`moveDown`, `down${bi}`, state?.delays?.[0]?.[bi]);
                    addMotion(`awake`, `down${bi}`, state?.delays?.[0]?.[bi]);
                    addMotion(`stop`, `down${bi}`, state?.delays?.[0]?.[bi]);
                    break;
                case 1:
                    output.push(`${outDir}up${bi}.png`);
                    addMotion(`moveUp`, `up${bi}`, state?.delays?.[0]?.[bi]);
                    break;
                case 2:
                    output.push(`${outDir}right${bi}.png`);
                    addMotion(`moveRight`, `right${bi}`, state?.delays?.[0]?.[bi]);
                    addMotion(`moveUpRight`, `right${bi}`, state?.delays?.[0]?.[bi]);
                    addMotion(`moveDownRight`, `right${bi}`, state?.delays?.[0]?.[bi]);
                    break;
                case 3:
                    output.push(`${outDir}left${bi}.png`);
                    addMotion(`moveLeft`, `left${bi}`, state?.delays?.[0]?.[bi]);
                    addMotion(`moveUpLeft`, `left${bi}`, state?.delays?.[0]?.[bi]);
                    addMotion(`moveDownLeft`, `left${bi}`, state?.delays?.[0]?.[bi]);
                    break;
                default:
                    throw new Error(`counting to ${fi} is hard :(`);
            }

            const o = output[0];
            const stream = canvas.createPNGStream();
            const out = fs.createWriteStream(o);

            stream.pipe(out);
            await new Promise(r => out.on('finish', () => r(0)));
            for (let io = 1; io < output.length; io++)
                fs.cpSync(o, output[io]);
        }
    }

    const deadFrames = getFrames(`${name}_dead.png`);
    for (let bi = 0; bi < deadFrames.length; bi++) {
        var canvas = Canvas.createCanvas(totalSize.x * resize, totalSize.y * resize);
        const ctx = canvas.getContext('2d');
        ctx.imageSmoothingEnabled = false;

        if (deadFrames[bi])
            deadFrames[bi](ctx);
        //let trim = trimCanvas(canvas);
        //if (trim)
        //    canvas = trim.canvas;

        const o = `${outDir}wait${bi}.png`;
        addMotion(`wait`, `wait${bi}`, state?.delays?.[0]?.[bi]);

        const stream = canvas.createPNGStream();
        const out = fs.createWriteStream(o);

        stream.pipe(out);
        await new Promise(r => out.on('finish', () => r(0)));
    }

    // gonna extend the death a bit
    let stopMotion = skin['motion-params'].motion.find(x => x['$'].state === `stop`);
    let longer = [];
    // @ts-ignore
    for (let i = 0; i < 7; i++)longer = longer.concat(stopMotion.item);
    // @ts-ignore
    stopMotion.item = longer;

    var builder = new xml2js.Builder();
    var xml = builder.buildObject(skin);
    //console.debug(xml);
    fs.writeFileSync(skinXml, xml)

    console.log(`done!`);
}

make(require(`./meta.json`))

